//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "jit_scalar.h"
#include "jit.h"
#include "x86disasm.h"
#include "log.h"
#include <xmmintrin.h>
#include <float.h>

extern const unsigned char g_VDDisasmX86[];

VDShaderScalarJIT::VDShaderScalarJIT()
	: mpShader(NULL)
{
}

VDShaderScalarJIT::~VDShaderScalarJIT() {
	if (mpShader) {
		mAllocator.Free(mpShader, mShaderLen);
		mpShader = NULL;
	}
}

void VDShaderScalarJIT::LoadSamplers(const VDShaderInterpreterSampler *samplers, int numSamplers) {
	mpSamplers = samplers;
}

namespace {
	uint32 SampleTexturePoint(VDShaderInterpreterSampler& samp, float u, float v) {
		if (!samp.mbClampU)
			u -= floorf(u);

		if (!samp.mbClampV)
			v -= floorf(v);

		u *= samp.mFWidth;
		v *= samp.mFHeight;

		int iu = (int)u;
		int iv = (int)v;

		if (samp.mbClampU && (unsigned)iu > samp.mULimit)
			iu = ~(iu >> 31) & samp.mULimit;
		if (samp.mbClampV && (unsigned)iv > samp.mVLimit)
			iv = ~(iv >> 31) & samp.mVLimit;

		return ((const uint32 *)((const char *)samp.mpSrc + samp.mPitch*iv))[iu];
	}

	uint32 SampleTextureBilinear(VDShaderInterpreterSampler& samp, float u, float v) {
		if (!samp.mbClampU)
			u -= floorf(u);

		if (!samp.mbClampV)
			v -= floorf(v);

		u *= samp.mFWidth;
		v *= samp.mFHeight;

		int iu = (int)u;
		int iv = (int)v;

		if (samp.mbClampU && (unsigned)iu > samp.mULimit)
			iu = ~(iu >> 31) & samp.mULimit;
		if (samp.mbClampV && (unsigned)iv > samp.mVLimit)
			iv = ~(iv >> 31) & samp.mVLimit;

		iu -= 32;
		iv -= 32;

		const uint32 *row0 = (const uint32 *)((const char *)samp.mpSrc + samp.mPitch*(iv >> 6)) + (iu >> 6);
		const uint32 *row1 = (const uint32 *)((const char *)row0 + samp.mPitch);
		uint32 px0 = row0[0];
		uint32 px1 = row0[1];
		uint32 px2 = row1[0];
		uint32 px3 = row1[1];

		uint32 px0_rb = px0 & 0xff00ff;
		uint32 px1_rb = px1 & 0xff00ff;
		uint32 px2_rb = px2 & 0xff00ff;
		uint32 px3_rb = px3 & 0xff00ff;
		uint32 px0_ag = px0 & 0xff00ff00;
		uint32 px1_ag = px1 & 0xff00ff00;
		uint32 px2_ag = px2 & 0xff00ff00;
		uint32 px3_ag = px3 & 0xff00ff00;

		sint32 uf = iu & 63;
		sint32 vf = iv & 63;

		uint32 px01_rb = (px0_rb + (((sint32)(px1_rb - px0_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px23_rb = (px2_rb + (((sint32)(px3_rb - px2_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px01_ag = (px0_ag + ((sint32)((px1_ag >> 6) - (px0_ag >> 6))*uf + 0x800080)) & 0xff00ff00;
		uint32 px23_ag = (px2_ag + ((sint32)((px3_ag >> 6) - (px2_ag >> 6))*uf + 0x800080)) & 0xff00ff00;

		uint32 px_rb = (px01_rb + (((sint32)(px23_rb - px01_rb)*vf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px_ag = (px01_ag + ((sint32)((px23_ag >> 6) - (px01_ag >> 6))*vf + 0x800080)) & 0xff00ff00;

		return px_rb + px_ag;
	}

	float Frac(float f) {
		return f - floorf(f);
	}

	float Log(float f) {
		float v = fabsf(f);

		if (v == 0.0f)
			v = -FLT_MAX;
		else
			v = logf(v) * 1.4426950408889634073599246810019f;

		return v;
	}

	float Exp(float f) {
		return expf(f * 0.69314718055994530941723212145818f);
	}
}

bool VDShaderScalarJIT::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	if (mpShader) {
		mAllocator.Free(mpShader, mShaderLen);
		mpShader = NULL;
	}

	// compile shader to scalar form
	mScalarForm.Clear();

	VDShaderScalarTransformer xform(mScalarForm);
	if (!xform.LoadShader(func, pElements, numElements, out))
		return false;

	using namespace x86;
	DynamicContext ctx;

	BasicBlock *bbent = ctx.GetEntryBlock();
	BasicBlock *bbexit = ctx.GetExitBlock();

	int numConstants = mScalarForm.mConstants.size();
	const float *pConstants = mScalarForm.mConstants.data();

	int numTemps = mScalarForm.mTempCount;
	int numInputs = mScalarForm.mInputCount;

	bbent->sub(esp, numTemps * 4 + numInputs * 8 + 16);

	static const float kDiv255 = 1.0f / 255.0f;

	bbent->mov(ebp, dword_ptr[esp + numTemps*4 + numInputs*8 + 44]);
	bbent->mov(ebx, dword_ptr[esp + numTemps*4 + numInputs*8 + 40]);

	for(int i=0; i<numInputs; ++i) {
		bbent->mov(eax, dword_ptr [ebx + 4*i]);
		bbent->mov(dword_ptr [esp + numTemps*4 + 16 + 8*i], eax);
		bbent->mov(eax, dword_ptr [ebx + 4*i + 16]);
		bbent->mov(dword_ptr [esp + numTemps*4 + 20 + 8*i], eax);
	}

	bbent->mov(ebx, dword_ptr[esp + numTemps*4 + numInputs*8 + 36]);

	struct Local {
		const dword_ref_t GetVariableSlot(int reg, int esp_offset = 0) {
			using namespace x86;

			switch(reg & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					return dword_ptr[esp + (reg - kVDShaderRegR0)*4 + mTempOffset + esp_offset];

				case kVDShaderRegC0:
					return dword_ptr[mpConstants + (reg - kVDShaderRegC0)];

				case kVDShaderRegV0:
					return dword_ptr[esp + (reg - kVDShaderRegV0)*4 + mInputOffset + esp_offset];

				default:
					VDASSERT(false);
					{
						static const float zero = 0.0f;
						return dword_ptr [&zero];
					}
			}
		}

		const float *mpConstants;
		int mNumConstants;
		ptrdiff_t mTempOffset;
		ptrdiff_t mInputOffset;
	} local = { pConstants, numConstants, 16, 16 + numTemps * 4 };

	BasicBlock *bb = ctx.CreateBasicBlock();
	BasicBlock *bbloop = bb;
	bbent->jmp(bbloop);

	int outputMappings[4] = { 0, 0, 0, 0 };

	VDShaderScalarForm::Insns::const_iterator it(mScalarForm.mInsns.begin()), itEnd(mScalarForm.mInsns.end());
	for(; it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;
		if (!insn.op)
			break;

		if (insn.op == kVDShaderInterpOp_beginloop) {
			bbloop = ctx.CreateBasicBlock();
			bb->jmp(bbloop);
			bb = bbloop;
			continue;
		}

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegO0) {
			VDASSERT(insn.op == kVDShaderInterpOp_mov);
			VDASSERT(insn.src1);

			outputMappings[insn.dst & 3] = insn.src1;
			continue;
		}

		VDASSERT(!insn.dst || (insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0);
		switch(insn.op) {
			case kVDShaderInterpOp_texld:
				bb->push(local.GetVariableSlot(insn.src2, 0));
				bb->push(local.GetVariableSlot(insn.src1, 4));

				bb->push((sint32)mpSamplers);

				if (mpSamplers[0].mbBilinear)
					bb->call(ctx.CreateAbsTarget(SampleTextureBilinear));
				else
					bb->call(ctx.CreateAbsTarget(SampleTexturePoint));

				{
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jmp(bb2);
					bb = bb2;
				}

				bb->add(esp, 12);
				bb->mov(dword_ptr[esp + (insn.dst - kVDShaderRegR0)*4 + 16], eax);
				continue;
			case kVDShaderInterpOp_unpackr:
			case kVDShaderInterpOp_unpackg:
			case kVDShaderInterpOp_unpackb:
			case kVDShaderInterpOp_unpacka:
				VDASSERT((insn.src1 & kVDShaderRegTypeMask) == kVDShaderRegR0);

				bb->movzx(eax, byte_ptr[esp + (insn.src1 - kVDShaderRegR0)*4 + (insn.op - kVDShaderInterpOp_unpackb) + 16]);
				bb->mov(dword_ptr [esp], eax);
				bb->fild(dword_ptr [esp]);
				bb->fld(dword_ptr [&kDiv255]);
				bb->fmulp();
				bb->fstp(local.GetVariableSlot(insn.dst));
				continue;
		}

		if (insn.src1)
			bb->fld(local.GetVariableSlot(insn.src1));

		if (insn.src2)
			bb->fld(local.GetVariableSlot(insn.src2));

		if (insn.src3)
			bb->fld(local.GetVariableSlot(insn.src3));

		switch(insn.op) {
			case kVDShaderInterpOp_mov:
				break;
			case kVDShaderInterpOp_add:
				bb->faddp();
				break;
			case kVDShaderInterpOp_sub:
				bb->fsubp();
				break;
			case kVDShaderInterpOp_mul:
				bb->fmulp();
				break;
			case kVDShaderInterpOp_rsq:
				bb->fsqrt();
				bb->fld1();
				bb->fdivrp();
				break;
			case kVDShaderInterpOp_rcp:
				bb->fld1();
				bb->fdivrp();
				break;
			case kVDShaderInterpOp_frc:
				{
					bb->push(eax);
					bb->fstp(dword_ptr [esp]);
					BasicBlock *bbfrc = ctx.CreateAbsTarget(Frac);
					bb->call(bbfrc);
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jmp(bb2);
					bb = bb2;
					bb->pop(eax);
				}
				break;
			case kVDShaderInterpOp_neg:
				bb->fchs();
				break;
			case kVDShaderInterpOp_min:
				{
					bb->fcom();
					bb->fnstsw_ax();
					bb->test(ah, 0x41);
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jnp(bb2);		// C3,C0 = 00 or 11
					BasicBlock *bb3 = ctx.CreateBasicBlock();
					bb->jmp(bb3);
					bb = bb3;
					bb->fxch(st1);
					bb->jmp(bb2);
					bb = bb2;
					bb->fstp(st1);
				}
				break;
			case kVDShaderInterpOp_max:
				{
					bb->fcom();
					bb->fnstsw_ax();
					bb->test(ah, 0x03);
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jp(bb2);		// C2,C0 = 01 or 10
					BasicBlock *bb3 = ctx.CreateBasicBlock();
					bb->jmp(bb3);
					bb = bb3;
					bb->fxch(st1);
					bb->jmp(bb2);
					bb = bb2;
					bb->fstp(st1);
				}
				break;
			case kVDShaderInterpOp_abs:
				bb->fabs();
				break;
			case kVDShaderInterpOp_log:
				{
					bb->push(eax);
					bb->fstp(dword_ptr [esp]);
					BasicBlock *bblog = ctx.CreateAbsTarget(Log);
					bb->call(bblog);
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jmp(bb2);
					bb = bb2;
					bb->pop(eax);
				}
				break;
			case kVDShaderInterpOp_exp:
				{
					bb->push(eax);
					bb->fstp(dword_ptr [esp]);
					BasicBlock *bbexp = ctx.CreateAbsTarget(Exp);
					bb->call(bbexp);
					BasicBlock *bb2 = ctx.CreateBasicBlock();
					bb->jmp(bb2);
					bb = bb2;
					bb->pop(eax);
				}
				break;
			case kVDShaderInterpOp_cmp:
				bb->fstp(dword_ptr [esp+8]);
				bb->fstp(dword_ptr [esp+4]);
				bb->fstp(dword_ptr [esp]);
				bb->mov(eax, dword_ptr [esp]);
				bb->cmp(eax, 0x80000001);
				bb->sbb(eax, eax);
				bb->fld(dword_ptr [esp+eax*4+8]);
				break;
		}

		if (insn.dst)
			bb->fstp(local.GetVariableSlot(insn.dst));
	}

	for(int i=0; i<4; ++i) {
		int r = outputMappings[i];

		if (r)
			bb->fld(local.GetVariableSlot(r));
		else
			bb->fldz();

		static const float kOffset = 12582912.0f;
		static const float kScale = 255.0f;
		bb->fmul(dword_ptr [&kScale]);
		bb->fadd(dword_ptr [&kOffset]);
		bb->fstp(dword_ptr [esp]);
		bb->mov(eax, dword_ptr [esp]);
		bb->sub(eax, 0x4b400000);
		bb->sbb(esi, esi);
		bb->or(eax, esi);
		bb->xor(esi, eax);
		bb->mov(eax, 255);
		bb->sub(eax, esi);
		bb->cdq();
		bb->and(edx, eax);
		bb->add(edx, esi);

		static const int kShifts[] = { 16, 8, 0, 24 };
		int shift = kShifts[i];

		if (shift)
			bb->shl(edx, shift);

		if (i)
			bb->add(edi, edx);
		else
			bb->mov(edi, edx);
	}

	bb->mov(dword_ptr [ebx], edi);
	bb->add(ebx, 4);
	bb->sub(ebp, 1);
	bb->jnz(bbloop);

	BasicBlock *bboutro = ctx.CreateBasicBlock();
	bb->jmp(bboutro);

	bboutro->add(esp, numTemps*4 + numInputs*8 + 16);
	bboutro->jmp(bbexit);

	ctx.Generate();

	uint32 codelen = ctx.GetCodeLen();
	mpShader = (Shader)mAllocator.Allocate(codelen);
	if (mpShader) {
		ctx.Copy(mpShader);
		mAllocator.EndUpdate(mpShader, codelen);

		if (out) {
			out->WriteLogOutput("//\n");
			out->WriteLogOutput("// Generated by VirtualDub scalar x86 JIT Compiler\n");
			out->WriteLogOutput("//\n");

			VDDisassemblyContext ctx;
			ctx.Init(g_VDDisasmX86);

			const unsigned char *sbase = (const unsigned char *)mpShader;
			const unsigned char *s = sbase;
			const unsigned char *send = sbase + codelen;

			while(s != send) {
				int count;
				const char *text = ctx.Disassemble(s, send - s, count);

				VDShaderWriteLogOutputF(*out, "%4x: %s\n", s - sbase, text);
				s += count;
			}
		}
	}

	return true;
}
