//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef PARSER_H
#define PARSER_H

#ifdef _MSC_VER
	#pragma once
#endif

#include <hash_map>
#include <vd2/system/refcount.h>
#include "effect.h"
#include "lexer.h"

namespace nsVDShaderParser {
	enum VDShaderParserType {
		kTypeInvalid,
		kTypeVoid,
		kTypeVertexShader,
		kTypePixelShader,
		kTypeFloat,
		kTypeFloat2,
		kTypeFloat3,
		kTypeFloat4,
		kTypeFloat2x2,
		kTypeFloat2x3,
		kTypeFloat2x4,
		kTypeFloat3x2,
		kTypeFloat3x3,
		kTypeFloat3x4,
		kTypeFloat4x2,
		kTypeFloat4x3,
		kTypeFloat4x4,
		kTypeSampler,
		kTypeBool,
		kTypeBool2,
		kTypeBool3,
		kTypeBool4,
	};

	const int kRowCountForType[]={
		0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 0, 1, 1, 1, 1
	};
	
	const int kColCountForType[]={
		0, 0, 0, 0, 1, 2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4, 0, 1, 2, 3, 4 
	};

	const int kColWriteMaskForType[]={
		0, 0, 0, 0, 1, 3, 7, 15, 3, 7, 15, 3, 7, 15, 3, 7, 15, 0, 1, 3, 7, 15
	};

	const unsigned kInsnComponentMasks[]={
		0x00000000,		// nop
		0x0000000F,		// abs
		0x000000FF,		// add
		0x00000FFF,		// cmp
		0x00000077,		// crs
		0x00000133,		// dp2add
		0x00000077,		// dp3
		0x000000FF,		// dp4
		0x00000001,		// exp
		0x00000001,		// frc
		0x00000001,		// log
		0x00000FFF,		// lrp
		0x00000077,		// m3x2
		0x00000077,		// m3x3
		0x00000077,		// m3x4
		0x000000FF,		// m4x3
		0x000000FF,		// m4x4
		0x00000FFF,		// mad
		0x000000FF,		// max
		0x000000FF,		// min
		0x0000000F,		// mov
		0x000000FF,		// mul
		0x000000FF,		// nrm
		0x00000011,		// pow
		0x00000001,		// rcp
		0x00000001,		// rsq
		0x00000FF1,		// sincos
		0x000000FF,		// sub
		0x0000000F,		// texkill
		0x000000F3,		// texld
		0x000000FF,		// texldb
		0x000000FF,		// texldp
	};
}

class VDShaderParserFunction;

struct VDShaderParserVariable : public VDShaderVariable {
	bool	mbIsConst;

	union {
		VDShaderConstant	m[4];
		int		mShader;
	};

	VDShaderParserVariable() {
		mbUniform = false;
	}
};

class VDShaderParserFunction : public VDShaderFunction {
public:
	VDShaderParserFunction();
	~VDShaderParserFunction();

	int AllocateParamRegisters(int rows);
	int AllocateRegisters(int rows);
	int AddConstantZero();
	int AddConstant(const VDShaderConstant&);
	int PromoteToConstantRegisters(VDShaderParserVariable& var);

	int AllocateSampler(int globalSamplerIndex);

	void mov(int dst, unsigned char mask, int src1);
	void mov_swizzle(int dst, unsigned char mask, int src1, unsigned char swizzle);
	void mov_mods(int dst, unsigned char mask, int src1, unsigned char swizzle, unsigned char modifiers);
	void abs(int dst, unsigned char mask, int src1);
	void add(int dst, unsigned char mask, int src1, int src2);
	void add_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods);
	void cmp(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, int src3);
	void cmp_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods, int src3, unsigned char src3swizzle, unsigned char src3mods);
	void crs(int dst, unsigned char mask, int src1, int src2);
	void sub(int dst, unsigned char mask, int src1, int src2);
	void mul(int dst, unsigned char mask, int src1, int src2);
	void mul_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle);
	void mad_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle);
	void mad_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods, int src3, unsigned char src3swizzle, unsigned char src3mods);
	void dp2add(int dst, unsigned char mask, int src1, int src2, int src3, unsigned char src3swizzle);
	void dp3(int dst, unsigned char mask, int src1, int src2);
	void dp4(int dst, unsigned char mask, int src1, int src2);
	void rsq(int dst, unsigned char mask, int src1);
	void rcp(int dst, unsigned char mask, int src1);
	void rcp(int dst, unsigned char mask, int src1, unsigned char src1component);
	void frc(int dst, unsigned char mask, int src1, unsigned char src1component);
	void lrp(int dst, unsigned char mask, int src1, int src2, int src3);
	void texld(int dst, unsigned char mask, int src1, unsigned char src1component, int src2);
	void min(int dst, unsigned char mask, int src1, int src2);
	void max(int dst, unsigned char mask, int src1, int src2);
	void nrm(int dst, unsigned char mask, int src1);
	void log(int dst, unsigned char mask, int src1);
	void log_swizzle(int dst, unsigned char mask, int src1, unsigned char src1component);
	void exp(int dst, unsigned char mask, int src1);
	void exp_swizzle(int dst, unsigned char mask, int src1, unsigned char src1component);

	void insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mods);
	void insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods);
	void insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods, int src3, unsigned char src3swizzle, unsigned char src3mods);
	void insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle);
	void insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle);

	bool Inline(VDShaderLexer& lexer, const VDShaderParserFunction& calledFunc, const VDShaderParserVariable *arguments, VDShaderVariable& returnValue);
	bool Optimize(VDShaderLexer& lexer, const VDShaderNameTable& nameTable);

	int mIndex;

	typedef stdext::hash_multimap<size_t, int> ConstantMap;
	ConstantMap mConstantMap;
};

class VDShaderParser {
public:
	VDShaderParser();
	~VDShaderParser();

	bool Parse(const char *src, size_t len, const char *context, bool optimizationEnabled, IVDShaderLogOutput *pOutput, VDShaderEffect **ppEffect);

	bool GetErrorInfo(VDShaderErrorInfo& errorInfo);

protected:
	bool TryParseGlobalStatement();
	bool ParseTechnique();
	bool ParsePass(VDShaderTechnique& tech);
	bool ParseFunction(int type, int name);
	bool ParseDeclarationList(int type);
	bool ParseDeclarationList(int type, int name);
	bool ParseBlock(bool& returnFound);
	bool ParseStatement(bool& returnFound);
	bool ParseReturnStatement();
	bool ParseExpression();
	bool ParseUnaryExpression(int tok);
	bool ParseValue(int tok);
	bool ParseSwizzle(int type, int maxComponents, unsigned char& swizzle, int& swizzleLength);
	bool ParseSamplerState();
	bool ParseCompileExpression();
	bool ParseCtorExpression(int type);
	int GetTypeFromIdent() const;
	int GetNameToken();
	const char *GetName(int name) const;
	const char *GetTypeName(int name) const;
	int FindLValue(int name, bool localOnly);
	int FindVariable(int name, bool localOnly);

	bool ConvertToType(VDShaderParserVariable& var, int type);
	bool ValidateFunctionAsPixelShader(const VDShaderParserFunction& func);

	int mNextGlobalConstantRegister;
	int mCurrentScopeLevel;
	int	mGlobalSamplers;
	VDShaderParserFunction *mpFunction;
	bool mbOptimizationEnabled;

	VDShaderLexer	mLexer;

	typedef std::vector<int> OpStack;
	OpStack mOpStack;

	VDShaderNameTable mNameTable;

	typedef std::vector<VDShaderParserVariable> VarHeap;
	VarHeap mVarHeap;

	typedef std::vector<float> Constants;
	Constants mConstants;

	typedef std::vector<VDShaderParserFunction *> Functions;
	Functions mFunctions;

	vdrefptr<VDShaderEffect> mpEffect;
};

const char *VDShaderGetTypeName(int type);

#endif
